package de.intevation.flys.exports;

import org.apache.log4j.Logger;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import org.w3c.dom.Document;

import de.intevation.artifacts.Artifact;

import de.intevation.flys.artifacts.WINFOArtifact;
import de.intevation.flys.artifacts.model.WQKms;


/**
 * An OutGenerator that generates discharge curves.
 *
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 */
public class LongitudinalSectionGenerator extends XYChartGenerator {

    /** The logger that is used in this generator.*/
    private static Logger logger =
        Logger.getLogger(LongitudinalSectionGenerator.class);


    public static final String LONGITUDINAL_SECTION_W =
        "longitudinal_section.w";

    public static final String LONGITUDINAL_SECTION_Q =
        "longitudinal_section.q";


    /** The storage for the W series to be drawn in this chart.*/
    protected XYSeriesCollection w;

    /** The storage for the Q series to be drawn in this chart.*/
    protected XYSeriesCollection q;


    public LongitudinalSectionGenerator() {
        super();

        this.w = new XYSeriesCollection();
        this.q = new XYSeriesCollection();
    }


    protected String getChartTitle() {
        // TODO i18n
        return "Wasserstand für Gewässer";
    }


    protected String getXAxisLabel() {
        return "km";
    }


    protected String getYAxisLabel() {
        return "W [NN + m]";
    }


    protected void addDatasets(JFreeChart chart) {
        XYPlot plot = (XYPlot) chart.getPlot();

        plot.setDataset(0, w);
        plot.setDataset(1, q);
    }


    protected void adjustAxes(XYPlot plot) {
        super.adjustAxes(plot);

        NumberAxis qAxis = new NumberAxis("Q [m³/s]");

        plot.setRangeAxis(2, qAxis);
        plot.mapDatasetToRangeAxis(1, 2);
    }


    public void doOut(Artifact artifact, String facet, Document attr) {
        logger.debug("LongitudinalSectionGenerator.doOut: " + facet);

        if (facet == null) {
            logger.error("No facet name for doOut(). No output generated!");
            return;
        }

        if (facet.equals(LONGITUDINAL_SECTION_W)) {
            doWOut(getWaterlevelData(artifact));
        }
        else if (facet.equals(LONGITUDINAL_SECTION_Q)) {
            doQOut(getWaterlevelData(artifact));
        }
        else {
            logger.warn("Unknown facet name: " + facet);
            return;
        }
    }


    /**
     * Returns the waterlevel data computed by the WINFOArtifact.
     *
     * @param artifact The WINFOArtifact.
     *
     * @return the computed waterlevel data.
     */
    protected WQKms[] getWaterlevelData(Artifact artifact) {
        WINFOArtifact winfoArtifact = (WINFOArtifact) artifact;
        WQKms[]       wqkms         = winfoArtifact.getWaterlevelData();

        logger.debug("Got " + wqkms.length + " WQKms objects.");

        return wqkms;
    }


    /**
     * Process the output for W facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     */
    protected void doWOut(WQKms[] wqkms) {
        logger.debug("LongitudinalSectionGenerator.doWOut");

        int idx = 0;
        for (WQKms tmp: wqkms) {
            XYSeries series = new XYSeries(getSeriesName(tmp, "w", idx++));

            double[] target = new double[3];
            int      size   = tmp.size();

            if (logger.isDebugEnabled()) {
                logger.debug("Generate series: " + series.getKey());

                logger.debug("Start km: " + tmp.getKms(0));
                logger.debug("End   km: " + tmp.getKms(size-1));
                logger.debug("Values  : " + size);
            }

            for (int i = 0; i < size; i++) {
                target = tmp.get(i, target);

                logger.debug("++ W Tuple: " + target[2] + " -> " + target[0]);
                series.add(target[2], target[0]);
            }

            w.addSeries(series);
        }
    }


    /**
     * Process the output for Q facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     */
    protected void doQOut(WQKms[] wqkms) {
        logger.debug("LongitudinalSectionGenerator.doQOut");

        int idx = 0;
        for (WQKms tmp: wqkms) {
            XYSeries series = new XYSeries(getSeriesName(tmp, "Q", idx++));

            double[] target = new double[3];
            int      size   = tmp.size();

            if (logger.isDebugEnabled()) {
                logger.debug("Generate series: " + series.getKey());

                logger.debug("Start km: " + tmp.getKms(0));
                logger.debug("End   km: " + tmp.getKms(size-1));
                logger.debug("Values  : " + size);
            }

            for (int i = 0; i < size; i++) {
                target = tmp.get(i, target);

                logger.debug("++ Q Tuple: " + target[2] + " -> " + target[1]);
                series.add(target[2], target[1]);
            }

            q.addSeries(series);
        }
    }


    protected String getSeriesName(WQKms wqkms, String prefix, int idx) {
        return prefix + "-" + idx;
    }
}
// vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
